#include <stdio.h>

#ifdef _WIN32
#  include <windows.h>
#elif (defined(__unix__) || defined(__unix) || (defined(__APPLE__) && defined(__MACH__)))
#  include <pthread.h>
#  include <unistd.h>
#else
#  error Unsupported plattform
#endif

#include "bakery.h"

#define ITERATIONS   30
#define LOCK_MSECS    5

static struct critical_section_t *critical_section;

static unsigned int value;

#ifdef _WIN32
DWORD WINAPI thread_func(LPVOID arg)
#else
void *	   thread_func (void *arg)
#endif
{
	unsigned long id = (unsigned long)arg;
	unsigned int i;

	for (i = 0; i < ITERATIONS; ++i) {
		My_EnterCriticalSection(id, critical_section);

		/*-- validation start --*/
		value = id;
#ifdef _WIN32
		Sleep(LOCK_MSECS);
#else
		usleep(LOCK_MSECS * 1000);
#endif
		if (value != id)
			printf("Thread %lu: invalid state! (i: %u)\n", id, i);
		else
			printf("Thread %lu: ok! (i: %u)\n", id, i);
		/*-- validation end --*/

		My_LeaveCriticalSection(id, critical_section);
	}

	return 0;
}

#ifdef _WIN32
int wmain()
#else
int main()
#endif
{
	My_InitializeCriticalSection(&critical_section);

#ifdef _WIN32
	HANDLE th1, th2, th3;

	th1 = CreateThread(NULL, 0, thread_func, (void*)1, 0, NULL);
	th2 = CreateThread(NULL, 0, thread_func, (void*)2, 0, NULL);
	th3 = CreateThread(NULL, 0, thread_func, (void*)3, 0, NULL);

	WaitForSingleObject(th1, INFINITE);
	WaitForSingleObject(th2, INFINITE);
	WaitForSingleObject(th3, INFINITE);
#else
	pthread_t th1, th2, th3;

	pthread_create(&th1, NULL, thread_func, (void*)1);
	pthread_create(&th2, NULL, thread_func, (void*)2);
	pthread_create(&th3, NULL, thread_func, (void*)3);

	pthread_join(th1, NULL);
	pthread_join(th2, NULL);
	pthread_join(th3, NULL);
#endif

	My_DeleteCriticalSection(critical_section);
	return 0;
}

